@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -45,6 +45,7 @@
 #define NV_CE_NUM_FBPCE                   4
 #define NV_CE_NUM_PCES_NO_LINK_CASE       12
 #define NV_CE_MAX_PCE_PER_GRCE            2
+#define NV_CE_HSHUBNVL_ID_0               2
 
 /*
  * Table for setting the PCE2LCE mapping for WAR configs that cannot be implemented
@@ -567,6 +568,61 @@ kceMapPceLceForGRCE_GH100
     }
 }
 
+/**
+ * @brief This function performs PCE-LCE mappings in the CC case where
+ *        a 1-1 mapping is applied.
+ *
+ * @param[in]   pGpu                        OBJGPU pointer
+ * @param[in]   pKCe                         KernelCE pointer
+ * @param[in]   pceAvailableMaskPerHshub    Pointer to CEs available per HSHUB
+ * @param[out]  pLocalPceLceMap             Pointer to PCE-LCE array
+ * @param[out]  pLocalExposeCeMask          Pointer to LCE Mask
+ */
+static void
+kceMapPceLceForCC
+(
+    OBJGPU  *pGpu,
+    KernelCE *pKCe,
+    NvU32   *pceAvailableMaskPerHshub,
+    NvU32   *pLocalPceLceMap,
+    NvU32   *pLocalGrceMap,
+    NvU32   *pLocalExposeCeMask
+)
+{
+    NvU32 pceIndex    = 0;
+    NvU32 maxLceIdx, lceMask, lceIndex, grceIdx;
+
+    // Apply 1-1 mapping for async LCEs
+    lceMask = NV_CE_MAX_LCE_MASK & (~kceGetGrceSupportedLceMask_HAL(pGpu, pKCe));
+    maxLceIdx = lceMask;
+    HIGHESTBITIDX_32(maxLceIdx);
+    *pLocalExposeCeMask |= lceMask;
+    lceIndex = CE_GET_LOWEST_AVAILABLE_IDX(lceMask);
+    for(; lceIndex <= maxLceIdx; lceIndex++)
+    {
+        pLocalPceLceMap[pceIndex] = lceIndex;
+        lceMask &= (~(NVBIT32(lceIndex)));
+        pceIndex++;
+    }
+
+
+    // Map GRCEs as non sharing. At this point, no PCEs have been mapped
+    lceMask = kceGetGrceSupportedLceMask_HAL(pGpu, pKCe);
+    *pLocalExposeCeMask |= lceMask;
+    for (grceIdx = 0; grceIdx < NV_CE_MAX_GRCE; grceIdx++)
+    {
+        lceIndex = CE_GET_LOWEST_AVAILABLE_IDX(lceMask);
+        if (((NVBIT32(lceIndex) & NV_CE_MAX_LCE_MASK) != 0) && (pceIndex < kceGetPce2lceConfigSize1_HAL(pKCe)))
+        {
+            pLocalPceLceMap[pceIndex] = lceIndex;
+            lceMask &= (~(NVBIT32(lceIndex)));
+            pLocalGrceMap[grceIdx] = DRF_NUM(_CE, _GRCE_CONFIG, _SHARED, 0) |
+                                     DRF_DEF(_CE, _GRCE_CONFIG, _SHARED_LCE, _NONE);
+            pceIndex++;
+        }
+    }
+}
+
 /**
  * @brief This function assigns PCE-LCE mappings for NVLink peers
  *        Based on HSHUBs that the links associated with a peer connect to,
@@ -891,6 +947,14 @@ kceGetMappings_GH100
         
     }
 
+    // In CC case a 1-1 mapping should be applied and other mappings are not required
+    if (gpuIsCCFeatureEnabled(pGpu))
+    {
+        kceMapPceLceForCC(pGpu, pKCe, pTopoParams->pceAvailableMaskPerHshub,
+                          pLocalPceLceMap, pLocalGrceMap, pExposeCeMask);
+        goto returnSuccess;
+    }
+
     //Prepare the per-HSHUB/FBHUB available PCE mask
     kceGetAvailableHubPceMask(pGpu, pKCe, pTopoParams);
 
@@ -929,5 +993,255 @@ kceGetMappings_GH100
     }
 
     NV_PRINTF(LEVEL_INFO, "status = %d, statusC2C = %d\n", status, statusC2C);
+
+returnSuccess:
+    return NV_OK;
+}
+
+NV_STATUS kceGetP2PCes_GH100(KernelCE *pKCe, OBJGPU *pGpu, NvU32 gpuMask, NvU32 *nvlinkP2PCeMask)
+{
+    //
+    // Currently Bug 4103154 requires an updated algorithm described below
+    // in the else case to assign the proper LCE for the direct connected systems. 
+    //
+
+    NvU32         gpuCount       = gpumgrGetSubDeviceCount(gpuMask);
+    NvU32         minP2PLce      = (NV_CE_EVEN_ASYNC_LCE_MASK | NV_CE_ODD_ASYNC_LCE_MASK) & NV_CE_MAX_LCE_MASK;
+    NvU32         i;
+    KernelNvlink  *pKernelNvlink = GPU_GET_KERNEL_NVLINK(pGpu);
+    NvBool        bSwitchConfig  = NV_FALSE;
+
+    if (pKernelNvlink == NULL)
+    {
+        return NV_WARN_NOTHING_TO_DO;
+    }
+
+    bSwitchConfig = knvlinkIsGpuConnectedToNvswitch(pGpu, pKernelNvlink);
+
+    LOWESTBITIDX_32(minP2PLce);
+    *nvlinkP2PCeMask  = 0;
+
+    if ((gpuCount == 1) && !bSwitchConfig)
+    {
+        *nvlinkP2PCeMask |= NVBIT(minP2PLce);
+        for (i = minP2PLce; i < gpuGetNumCEs(pGpu); i++)
+        {
+            *nvlinkP2PCeMask |= NVBIT(i);
+
+        }
+    }
+    //
+    // For cases where we have an nvswitch connected, we will assign
+    // the LCE with max PCEs
+    //
+    else if (bSwitchConfig)
+    {
+        KernelCE     *pKCeMaxPces   = NULL;
+        KernelCE     *pTargetCe     = NULL;
+        KernelCE     *pKCeLoop      = NULL;
+        NvU32        gpuInstance    = 0;
+        NvU32         maxPces       = 0;
+
+        KCE_ITER_ALL_BEGIN(pGpu, pKCeLoop, minP2PLce)
+
+        if (pKCeLoop->bStubbed)
+        {
+            continue;
+        }
+
+        NV2080_CTRL_CE_GET_CE_PCE_MASK_PARAMS params = {0};
+
+        // We will use LCE with most PCEs
+        params.ceEngineType = NV2080_ENGINE_TYPE_COPY(pKCeLoop->publicID);
+        NV_STATUS rmStatus = knvlinkExecGspRmRpc(pGpu, pKernelNvlink,
+                                                 NV2080_CTRL_CMD_CE_GET_CE_PCE_MASK,
+                                                 (void *)&params, sizeof(params));
+        NV_ASSERT_OK_OR_RETURN(rmStatus);
+        NvU32 numPces = nvPopCount32(params.pceMask);
+
+        if (numPces > maxPces)
+        {
+            pKCeMaxPces = pKCeLoop;
+            maxPces = numPces;
+        }
+        KCE_ITER_END
+
+        // For GPU connected to nvswitch, optimal LCE is always LCE with max PCE
+
+        if (pKCeMaxPces != NULL)
+        {
+            pTargetCe = pKCeMaxPces;
+        }
+
+        if (pTargetCe != NULL)
+        {
+            // assign LCE to peer
+            if (pTargetCe->nvlinkPeerMask == 0)
+            {
+                pTargetCe->nvlinkPeerMask = NVBIT(gpuInstance);
+            }
+
+            NV_PRINTF(LEVEL_INFO,
+                      "GPU %d Assigning Peer %d to LCE %d\n",
+                      gpuGetInstance(pGpu), gpuInstance,
+                      pTargetCe->publicID);
+
+            *nvlinkP2PCeMask = NVBIT(pTargetCe->publicID);
+        }
+    }
+    else if (gpuCount > 2)
+    {
+        // if gpuCount > 2, this is an invalid request. Print warning and return NV_OK
+        NV_PRINTF(LEVEL_INFO, "GPU %d invalid request for gpuCount %d\n", gpuGetInstance(pGpu), gpuCount);
+        return NV_ERR_INVALID_STATE;
+    }
+    else
+    {
+        OBJGPU       *pRemoteGpu        = NULL;
+        KernelCE     *pKCeLoop          = NULL;
+        NvU32         peerLinkMask      = 0;
+        NvU32         gpuInstance       = 0;
+        NvU32         phyLinkId, status, targetPceMask, numPces;
+
+        //
+        // The LCE returned should be the LCE which has the most PCEs mapped
+        // on the given HSHUB. This HSHUB should be determined by
+        // tracking where the majority of links are connected.
+        //
+        NvU32     linksPerHshub[NV_CE_MAX_HSHUBS] = {0};
+        NvU32     maxLinksConnectedHshub = 0;
+        NvU32     maxConnectedHshubId = NV_CE_MAX_HSHUBS;
+        NvU32     lceAssignedMask = 0;
+        KernelCE *maxLcePerHshub[NV_CE_MAX_HSHUBS] = {0};
+
+        NV2080_CTRL_INTERNAL_HSHUB_GET_HSHUB_ID_FOR_LINKS_PARAMS params;
+
+        if (pKernelNvlink != NULL)
+        {
+            // Get the remote GPU
+            while ((pRemoteGpu = gpumgrGetNextGpu(gpuMask, &gpuInstance)) != NULL)
+            {
+                if (pRemoteGpu != pGpu)
+                    break;
+            }
+
+            NV_ASSERT_OR_RETURN(pRemoteGpu != NULL, NV_ERR_INVALID_STATE);
+            gpuInstance = gpuGetInstance(pRemoteGpu);
+
+            peerLinkMask = knvlinkGetLinkMaskToPeer(pGpu, pKernelNvlink, pRemoteGpu);
+        }
+
+        portMemSet(&params, 0, sizeof(params));
+        params.linkMask = peerLinkMask;
+
+        status = knvlinkExecGspRmRpc(pGpu, pKernelNvlink,
+                                     NV2080_CTRL_CMD_INTERNAL_HSHUB_GET_HSHUB_ID_FOR_LINKS,
+                                     (void *)&params, sizeof(params));
+        NV_ASSERT_OK_OR_RETURN(status);
+
+        FOR_EACH_INDEX_IN_MASK(32, phyLinkId, peerLinkMask)
+        {
+            NvU32 hshubId = params.hshubIds[phyLinkId];
+            linksPerHshub[hshubId]++;
+
+            if (linksPerHshub[hshubId] > maxLinksConnectedHshub)
+            {
+                maxLinksConnectedHshub = linksPerHshub[hshubId];
+                maxConnectedHshubId = hshubId;
+            }
+        }
+        FOR_EACH_INDEX_IN_MASK_END;
+
+        //
+        // Iterate through all Async LCEs to track which HSHUB should
+        // be using which LCE. This is decided based on the majority. If
+        // there is a tie, then LCE with the lower index is preferred.
+        //
+        KCE_ITER_ALL_BEGIN(pGpu, pKCeLoop, minP2PLce)
+            NvU32 localMaxPcePerHshub = 0;
+            KernelCE *localMaxLcePerHshub;
+            NvU32 localMaxHshub = NV_CE_MAX_HSHUBS;
+
+            // if LCE is stubbed or LCE is already assigned to another peer
+            if (pKCeLoop->bStubbed)
+            {
+                continue;
+            }
+
+            // LCE is already assigned to this peer
+            if ((pKCeLoop->nvlinkPeerMask & NVBIT(gpuInstance)) != 0)
+            {
+                maxLcePerHshub[maxConnectedHshubId] = pKCeLoop;
+                break;
+            }
+            // LCE is already assigned to another peer
+            else if (pKCeLoop->nvlinkPeerMask != 0)
+            {
+                continue;
+            }
+
+            NV2080_CTRL_CE_GET_CE_PCE_MASK_PARAMS params = {0};
+
+            params.ceEngineType = NV2080_ENGINE_TYPE_COPY(pKCeLoop->publicID);
+            status = knvlinkExecGspRmRpc(pGpu, pKernelNvlink,
+                                                     NV2080_CTRL_CMD_CE_GET_CE_PCE_MASK,
+                                                     (void *)&params, sizeof(params));
+            NV_ASSERT_OK_OR_RETURN(status);
+
+            //
+            // An LCE may be utilized across several HSHUBs. Loop through all HSHUBs
+            // in order to decide which HSHUB holds the majority of this specific LCE.
+            // To help with this, create a mask of PCEs only on the HSHUB which the peer
+            // is most connected to by shifting the HSHUB PCE mask
+            //
+
+            for (i = NV_CE_HSHUBNVL_ID_0; i < NV_CE_MAX_HSHUBS; i++)
+            {
+                targetPceMask = params.pceMask & ((NVBIT(NV_CE_PCE_PER_HSHUB) - 1) << ((i - NV_CE_HSHUBNVL_ID_0) * NV_CE_PCE_PER_HSHUB));
+                numPces = nvPopCount32(targetPceMask);
+                if (numPces > localMaxPcePerHshub && !(lceAssignedMask & NVBIT(pKCeLoop->publicID)))
+                {
+                    localMaxPcePerHshub = numPces;
+                    localMaxLcePerHshub = pKCeLoop;
+                    localMaxHshub = i;
+                }
+            }
+
+            if (localMaxHshub < NV_CE_MAX_HSHUBS)
+            {
+                maxLcePerHshub[localMaxHshub] = localMaxLcePerHshub;
+                lceAssignedMask |= NVBIT(localMaxLcePerHshub->publicID);
+            }
+
+        KCE_ITER_END
+
+        if (maxLcePerHshub[maxConnectedHshubId] != NULL)
+        {
+            NV_PRINTF(LEVEL_INFO,
+                      "GPU %d Assigning Peer %d to preferred LCE %d\n",
+                      gpuGetInstance(pGpu), gpuInstance,
+                      maxLcePerHshub[maxConnectedHshubId]->publicID);
+        }
+        else
+        {
+            //
+            // In the event that the preferred HSHUB's primary LCE is not available,
+            // use the first available LCE and set that index as the new preferred hshub
+            //
+            maxLcePerHshub[maxConnectedHshubId] = GPU_GET_KCE(pGpu, maxConnectedHshubId);
+
+            NV_PRINTF(LEVEL_INFO,
+                      "GPU %d Assigning Peer %d to first available LCE %d\n",
+                      gpuGetInstance(pGpu), gpuInstance,
+                      maxLcePerHshub[maxConnectedHshubId]->publicID);
+        }
+
+        if (maxConnectedHshubId < NV_CE_MAX_HSHUBS)
+        {
+            maxLcePerHshub[maxConnectedHshubId]->nvlinkPeerMask = NVBIT(gpuInstance);
+            *nvlinkP2PCeMask = NVBIT(maxLcePerHshub[maxConnectedHshubId]->publicID);
+        }
+    }
+
     return NV_OK;
 }
