@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -45,6 +45,9 @@
 #include "kernel/gpu/conf_compute/conf_compute.h"
 
 static NV_STATUS _kfifoGetCaps(OBJGPU *pGpu, NvU8 *pKfifoCaps);
+static NV_STATUS _kfifoDisableChannelsForKeyRotation(OBJGPU *pGpu, RmCtrlParams *pRmCtrlParams,
+                                                     NvBool bEnableAfterKeyRotation, NvBool bForceKeyRotation,
+                                                     NV2080_CTRL_FIFO_DISABLE_CHANNELS_PARAMS *pParams);
 
 /*!
  * @brief deviceCtrlCmdFifoGetChannelList
@@ -780,12 +783,10 @@ subdeviceCtrlCmdFifoDisableChannelsForKeyRotation_IMPL
 )
 {
     NV_STATUS       status        = NV_OK;
-    NV_STATUS       tmpStatus     = NV_OK;
     OBJGPU         *pGpu          = GPU_RES_GET_GPU(pSubdevice);
     CALL_CONTEXT   *pCallContext  = resservGetTlsCallContext();
     RmCtrlParams   *pRmCtrlParams = pCallContext->pControlParams;
-    NvU32           i;
-    KernelChannel  *pKernelChannel = NULL;
+    NvU32           i = 0;
 
     NV_CHECK_OR_RETURN(LEVEL_INFO,
         pDisableChannelParams->numChannels <= NV_ARRAY_ELEMENTS(pDisableChannelParams->hChannelList),
@@ -793,16 +794,86 @@ subdeviceCtrlCmdFifoDisableChannelsForKeyRotation_IMPL
     ct_assert(NV_ARRAY_ELEMENTS(pDisableChannelParams->hClientList) == \
               NV_ARRAY_ELEMENTS(pDisableChannelParams->hChannelList));
 
-    // Send RPC to handle message on Host-RM
-    if (IS_VIRTUAL(pGpu) || IS_GSP_CLIENT(pGpu))
+    NV2080_CTRL_FIFO_DISABLE_CHANNELS_PARAMS *pParams = NULL;
+    pParams = portMemAllocNonPaged(sizeof(NV2080_CTRL_FIFO_DISABLE_CHANNELS_PARAMS));
+    NV_ASSERT_OR_RETURN(pParams != NULL, NV_ERR_NO_MEMORY);
+    portMemSet(pParams, 0, sizeof(NV2080_CTRL_FIFO_DISABLE_CHANNELS_PARAMS));
+
+    for (i = 0; i < pDisableChannelParams->numChannels; i++)
     {
-        NV_RM_RPC_CONTROL(pGpu,
-                          pRmCtrlParams->hClient,
-                          pRmCtrlParams->hObject,
-                          pRmCtrlParams->cmd,
-                          pRmCtrlParams->pParams,
-                          pRmCtrlParams->paramsSize,
-                          status);
+        pParams->hClientList[i] = pDisableChannelParams->hClientList[i];
+        pParams->hChannelList[i] = pDisableChannelParams->hChannelList[i];
+    }
+    pParams->numChannels = pDisableChannelParams->numChannels;
+    status = _kfifoDisableChannelsForKeyRotation(pGpu, pRmCtrlParams, pDisableChannelParams->bEnableAfterKeyRotation,
+                                                 NV_FALSE, pParams);
+    portMemFree(pParams);
+    return status;
+}
+
+/**
+ * @brief This does the same thing as @ref subdeviceCtrlCmdFifoDisableChannelsForKeyRotation_IMPL
+ *        with the main difference being it operates on a single client and take a RO API lock.
+ */
+NV_STATUS
+subdeviceCtrlCmdFifoDisableChannelsForKeyRotationV2_IMPL
+(
+    Subdevice *pSubdevice,
+    NV2080_CTRL_FIFO_DISABLE_CHANNELS_FOR_KEY_ROTATION_V2_PARAMS *pDisableChannelParams
+)
+{
+    NV_STATUS       status        = NV_OK;
+    OBJGPU         *pGpu          = GPU_RES_GET_GPU(pSubdevice);
+    CALL_CONTEXT   *pCallContext  = resservGetTlsCallContext();
+    RmCtrlParams   *pRmCtrlParams = pCallContext->pControlParams;
+    NvU32           i = 0;
+
+    NV_CHECK_OR_RETURN(LEVEL_INFO,
+        pDisableChannelParams->numChannels <= NV_ARRAY_ELEMENTS(pDisableChannelParams->hChannelList),
+        NV_ERR_INVALID_ARGUMENT);
+
+    NV2080_CTRL_FIFO_DISABLE_CHANNELS_PARAMS *pParams = NULL;
+    pParams = portMemAllocNonPaged(sizeof(NV2080_CTRL_FIFO_DISABLE_CHANNELS_PARAMS));
+    NV_ASSERT_OR_RETURN(pParams != NULL, NV_ERR_NO_MEMORY);
+    portMemSet(pParams, 0, sizeof(NV2080_CTRL_FIFO_DISABLE_CHANNELS_PARAMS));
+
+    for (i = 0; i < pDisableChannelParams->numChannels; i++)
+    {
+        pParams->hClientList[i] = pRmCtrlParams->hClient;
+        pParams->hChannelList[i] = pDisableChannelParams->hChannelList[i];
+    }
+    pParams->numChannels = pDisableChannelParams->numChannels;
+    status = _kfifoDisableChannelsForKeyRotation(pGpu, pRmCtrlParams, pDisableChannelParams->bEnableAfterKeyRotation,
+                                                 NV_TRUE, pParams);
+    portMemFree(pParams);
+    return status;
+}
+
+static NV_STATUS
+_kfifoDisableChannelsForKeyRotation
+(
+    OBJGPU         *pGpu,
+    RmCtrlParams   *pRmCtrlParams,
+    NvBool          bEnableAfterKeyRotation,
+    NvBool          bForceKeyRotation,
+    NV2080_CTRL_FIFO_DISABLE_CHANNELS_PARAMS *pParams
+)
+{
+    NV_STATUS       status        = NV_OK;
+    NV_STATUS       tmpStatus     = NV_OK;
+    NvU32           i;
+    KernelChannel  *pKernelChannel = NULL;
+    RM_API         *pRmApi        = GPU_GET_PHYSICAL_RMAPI(pGpu);
+
+    // Send RPC to handle message on Host-RM
+    if (IS_GSP_CLIENT(pGpu))
+    {
+        status = pRmApi->Control(pRmApi,
+                                 pRmCtrlParams->hClient,
+                                 pRmCtrlParams->hObject,
+                                 pRmCtrlParams->cmd,
+                                 pRmCtrlParams->pParams,
+                                 pRmCtrlParams->paramsSize);
     }
     // Send internal control call to actually disable channels and preempt channels
     else
@@ -812,42 +883,87 @@ subdeviceCtrlCmdFifoDisableChannelsForKeyRotation_IMPL
     }
 
     // Loop through all the channels and mark them disabled
-    for (i = 0; i < pDisableChannelParams->numChannels; i++)
+    NvBool bFound = NV_FALSE;
+    NvU32 h2dKeyList[CC_KEYSPACE_TOTAL_SIZE];
+    NvU32 keyIndex = 0;
+    ConfidentialCompute *pConfCompute = GPU_GET_CONF_COMPUTE(pGpu);
+
+    for (i = 0; i < pParams->numChannels; i++)
     {
         RsClient              *pClient = NULL;
         tmpStatus = serverGetClientUnderLock(&g_resServ,
-                                          pDisableChannelParams->hClientList[i], &pClient);
+                                          pParams->hClientList[i], &pClient);
         if (tmpStatus != NV_OK)
         {
             status = tmpStatus;
-            NV_PRINTF(LEVEL_ERROR, "Failed to get client with hClient = 0x%x status = 0x%x\n", pDisableChannelParams->hClientList[i], status);
+            NV_PRINTF(LEVEL_ERROR, "Failed to get client with hClient = 0x%x status = 0x%x\n", pParams->hClientList[i], status);
             continue;
         }
         tmpStatus = CliGetKernelChannel(pClient,
-                                     pDisableChannelParams->hChannelList[i], &pKernelChannel);
+                                     pParams->hChannelList[i], &pKernelChannel);
         if (tmpStatus != NV_OK)
         {
             status = tmpStatus;
             NV_PRINTF(LEVEL_ERROR, "Failed to get channel with hclient = 0x%x hChannel = 0x%x status = 0x%x\n", 
-                                    pDisableChannelParams->hClientList[i], pDisableChannelParams->hChannelList[i], status);
+                                    pParams->hClientList[i], pParams->hChannelList[i], status);
             continue;
         }
         kchannelDisableForKeyRotation(pGpu, pKernelChannel, NV_TRUE);
-        kchannelEnableAfterKeyRotation(pGpu, pKernelChannel, pDisableChannelParams->bEnableAfterKeyRotation);
+        kchannelEnableAfterKeyRotation(pGpu, pKernelChannel, bEnableAfterKeyRotation);
+        if (IS_GSP_CLIENT(pGpu))
+        {
+            NvU32 h2dKey, d2hKey;
+            NV_ASSERT_OK_OR_RETURN(confComputeGetKeyPairByChannel_HAL(pGpu, pConfCompute, pKernelChannel,
+                                                                      &h2dKey, &d2hKey));
+            if (bForceKeyRotation)
+            {
+                //
+                // This loop doesn't need to execute in the first iteration of above loop
+                // since keyList is empty.
+                //
+                for (NvU32 j = 0; j < keyIndex; j++)
+                {
+                    if (h2dKeyList[j] == h2dKey)
+                    {
+                        bFound = NV_TRUE;
+                        break;
+                    }
+                }
+                if (!bFound)
+                {
+                    NV_ASSERT_OR_RETURN(keyIndex < CC_KEYSPACE_TOTAL_SIZE, NV_ERR_INVALID_STATE);
+                    h2dKeyList[keyIndex++] = h2dKey;
+                }
+                bFound = NV_FALSE;
+            }
+            else
+            {
+                KEY_ROTATION_STATUS state;
+                NV_ASSERT_OK_OR_RETURN(confComputeGetKeyRotationStatus(pConfCompute, h2dKey, &state));
+                if ((state == KEY_ROTATION_STATUS_PENDING) ||
+                    (state == KEY_ROTATION_STATUS_PENDING_TIMER_SUSPENDED))
+                {
+                    NV_ASSERT_OK_OR_RETURN(confComputeCheckAndPerformKeyRotation(pGpu, pConfCompute, h2dKey, d2hKey));
+                }
+            }
+        }
     }
 
-    if ((IS_VIRTUAL(pGpu) || IS_GSP_CLIENT(pGpu)) &&
-        (pKernelChannel != NULL))
+    if (IS_GSP_CLIENT(pGpu) && bForceKeyRotation)
     {
-        NvU32 h2dKey, d2hKey;
-        ConfidentialCompute *pConfCompute = GPU_GET_CONF_COMPUTE(pGpu);
-        NV_ASSERT_OK_OR_RETURN(confComputeGetKeyPairByChannel_HAL(pGpu, pConfCompute, pKernelChannel, &h2dKey, &d2hKey));
-        KEY_ROTATION_STATUS state;
-        NV_ASSERT_OK_OR_RETURN(confComputeGetKeyRotationStatus(pConfCompute, h2dKey, &state));
-        if (state == KEY_ROTATION_STATUS_PENDING)
+        for (NvU32 j = 0; j < keyIndex; j++)
         {
-            NV_ASSERT_OK_OR_RETURN(confComputeCheckAndScheduleKeyRotation(pGpu, pConfCompute, h2dKey, d2hKey));
+            NvU32 h2dKey, d2hKey;
+            confComputeGetKeyPairByKey(pConfCompute, h2dKeyList[j], &h2dKey, &d2hKey);
+            NV_PRINTF(LEVEL_INFO, "Forcing key rotation on h2dKey 0x%x\n", h2dKey);
+            status = confComputeForceKeyRotation(pGpu, pConfCompute, h2dKey, d2hKey);
+            if (status != NV_OK)
+            {
+                NV_PRINTF(LEVEL_ERROR, "Forced key rotation for key 0x%x failed\n", h2dKey);
+                return status;
+            }
         }
     }
     return status;
 }
+
